module Gerador where

import System.Random
import Graphics.Gloss

import Tetromino
import Graficos

-- =========================================================================
-- Esse arquivo define o gerador de tetrominoes, um gerador que cria a lista
-- pseudo embaralhada com os tetrominoes, na ordem que o jogador os recebera
-- =========================================================================

-- =========================================================================
-- Esse gerador vai manter uma lista com os tetrominoes que vao aparecer
-- para o jogador. A regra de geracao utilizada eh
data GeradorTetros = GeradorTetros { _rng        :: StdGen
                                   , _proxTetros :: [Tetromino]
                                   }

-- =========================================================================
-- Funcao que gera uma lista de tetrominoes que serao os proximos a aparecerem
-- para o jogador. A regra para geracao eh: pegue a lista com todos
-- os tetrominoes, embaralhe essa lista e adicione no fim da lista atual
geraProximos :: GeradorTetros -> GeradorTetros
geraProximos (GeradorTetros rng lista) = do
    let (novoRng, rng') = split rng      

    let novaLista       = quickShuffle todosTetrominoes rng' -- Gera uma nova lista usando rng'
    let listaNova       = lista ++ novaLista                 -- Concatena com a lista atual

    GeradorTetros novoRng listaNova -- Novo gerador com um novo rng e a nova lista

-- =========================================================================

consomeTetromino :: GeradorTetros -> (GeradorTetros, Tetromino)
consomeTetromino g@(GeradorTetros _ lista) = do
    let geradorAtual = if length lista == 0 then geraProximos g  else g -- Certifica que o gerador vai ter elementos para pegar

    let listaAtual   = _proxTetros geradorAtual -- Pega a lista de tetrominoes do gerador atual
    let tetroAtual   = head listaAtual          -- Pega o proximo tetromino
    let listaNova    = tail listaAtual          -- Remove o tetroAtual da listaAtual

    let geradorNovo  = GeradorTetros (_rng geradorAtual) listaNova -- Monta o gerador com a lista atualizada

    (geradorNovo, tetroAtual)

-- =========================================================================
-- Gera um booleano aleatorio. Sera util na hora de embaralhar a lista de
-- tetrominoes
embaralha :: StdGen -> (Bool, StdGen)
embaralha g = (b, g')
    where
        (val, g') = random g :: (Float, StdGen)
        b         = val < 0.5

-- Gera uma lista de booleanos que tem o mesmo comprimento da lista passada.
-- Essa lista de booleanos diz se cada elemento deve ser colocado antes ou 
-- depois do elemento base.
geraListaEmbaralhar :: [a] -> StdGen -> [Bool]
geraListaEmbaralhar [] _     = []
geraListaEmbaralhar (_:as) g = b:(geraListaEmbaralhar as g')
    where
        (b, g') = embaralha g

-- Eh semelhante ao quicksort, porem eh usado para embaralhar uma lista.
-- A funcao escolhe um elemento base, no caso o primeiro, e decide se cada
-- elemento deve ir antes ou depois dele. A funcao eh aplicada recursivamente
-- para os elementos que ficaram antes e depois do elemento base.
quickShuffle :: [a] -> StdGen -> [a]
quickShuffle [] _     = []
quickShuffle (a:as) g = (quickShuffle menores gMen) ++ [a] ++ (quickShuffle maiores gMai)
    where
        listaEmbaralhar  = geraListaEmbaralhar as gAtual                    -- Lista que diz quais elementos devem ir para maiores e quais para menores
        tuplasEmbaralhar = zip as listaEmbaralhar                           -- Tuplas que dizem quais elementos devem ir para maiores e quais para menores
        maiores          = [val | (val, b) <- tuplasEmbaralhar, b == True]  -- Lista com os elementos que ficarao depois de 'a'
        menores          = [val | (val, b) <- tuplasEmbaralhar, b == False] -- Lista com os elementos que ficarao antes de 'a'
        (gAtual, g')     = split g                                          -- gAtual eh usado decidir os maiores/menores com relacao a 'a'
        (gMai, gMen)     = split g'                                         -- gMai e gMen sao usados recursivamente para as listas com os maiores e menores

-- =========================================================================
-- Devolve os primeiros n tetrominoes do gerador
consultaN :: GeradorTetros -> Int -> (GeradorTetros, [Tetromino])
consultaN gerador n = do 
    let geradorAtual = validaGerador gerador n           -- Garante que a lista tera no minimo n elementos
    let consulta     = take n $ _proxTetros geradorAtual -- Pega uma os n primeiros, sem remover eles da lista atual

    (geradorAtual, consulta)

-- Caso a lista atual ja tenha n ou mais elementos, nao faz nada, caso contrario
-- adiciona elementos ate a lista ter n ou mais
validaGerador :: GeradorTetros -> Int -> GeradorTetros
validaGerador g@(GeradorTetros _ lista) n = 
    if   length lista >= n
    then g
    else validaGerador (geraProximos g) n

-- =========================================================================
-- Desenha os 6 primeiros tetrominoes do gerador
desenhaProxTetros :: GeradorTetros -> Float -> (GeradorTetros, Picture)
desenhaProxTetros gerador tamanhoBloco = do
    -- Pega os seis primeiros tetrominoes do gerador
    let (gerador', proxTetrominos) = consultaN gerador 6

    -- Desenha o background
    let background = montaRetangulo 0 0 (6*tamanhoBloco) (20*tamanhoBloco) 2 white black

    -- Desenha a lista dos proximos tetrominoes
    let proxTetrosGrafico = desenhaLista tamanhoBloco (-tamanhoBloco) tamanhoBloco proxTetrominos
     
    -- Desenha o titulo, i.e., a palavra "Proximos" acima a lista, junto com um fundo branco para
    -- ele
    let tituloBg      = montaRetangulo 0 0 (6*tamanhoBloco) (-(1.5*tamanhoBloco)) 2 white white
    let tituloGrafico = Translate 0 3 $ Scale (tamanhoBloco/100) (tamanhoBloco/100) $ text "Proximos"

    (gerador', Pictures [background, tituloBg, tituloGrafico, proxTetrosGrafico])

-- Desenha a lista de tetrominoes
desenhaLista :: Float -> Float -> Float -> [Tetromino] -> Picture
desenhaLista _ _ _ []     = Blank
desenhaLista x y tamanhoBloco (t:ts) = Pictures [atual, prox]
    where
        atual = desenhaLayout (_layout t) x y tamanhoBloco False False
        prox  = desenhaLista x y' tamanhoBloco ts
        y'    = y - (3*tamanhoBloco)

