module Jogo where

import Graphics.Gloss
import Graphics.Gloss.Interface.IO.Interact

import Gerador
import Tabuleiro
import Tetromino
import Matriz
import Graficos
import Vetor

-- =========================================================================
-- Esse arquivo possui as definicoes do jogo em si, i.e. as funcoes que juntam
-- tudo que foi desenvolvido para montar o jogo
-- =========================================================================

-- =========================================================================
-- IMPORTANTE: Apesar de nao ser o usual na matematica, o Tabuleiro do
-- do tetris sera programado pensando na origem como sendo o seu canto
-- superior esquerdo, sendo que o eixo X cresce para direita e o eixo Y
-- cresce para BAIXO. Esse eh o padrao em algumas outras engines de programacao
-- de jogos 2d, portanto, sera adotado aqui tambem.
-- =========================================================================

data Jogo = Jogo { _tabuleiro    :: Tabuleiro       -- Tabuleiro atual
                 , _hold         :: Maybe Tetromino -- Tetromino em hold atual
                 , _trocou       :: Bool            -- Diz se o jogador trocou o tetro atual com o hold
                 , _geradorTets  :: GeradorTetros   -- Gerador dos proximos tetrominoes
                 , _level        :: Int             -- Nivel atual do jogador
                 , _linhasFeitas :: Int             -- Quantidade de linhas feitas
                 , _pontuacao    :: Int             -- Pontuacao atual do jogador
                 , _combo        :: Int             -- Quantidade de combo atual
                 , _estaJogando  :: Bool            -- Diz se o jogo ja comecou ou nao
                 , _tamanhoBloco :: Float           -- Diz o tamanho que cada bloco dos tetrominos deve ser desenhado
                 , _gravidade    :: Float           -- Indica quantos quadrados o tetromino cai por frame
                 , _tParaCair    :: Float           -- Indica o tempo que demora para cair uma celula
                 , _tEsperando   :: Float           -- Quanto tempo faz que o jogo derrubou uma peca ou recebeu um input bem sucedido
                 }

-- =========================================================================
-- Altera o gerador do jogo atual
setGerador :: Jogo -> GeradorTetros -> Jogo
setGerador (Jogo t h troc _ lvl li po c estaJ tam grav tCair msEspand) gNovo = Jogo t h troc gNovo lvl li po c estaJ tam grav tCair msEspand

-- =========================================================================
-- Altera o gerador do jogo atual
setEstaJogando :: Jogo -> Bool -> Jogo
setEstaJogando (Jogo t h troc ger lvl li po c _ tam grav tCair msEspand) bNovo = Jogo t h troc ger lvl li po c bNovo tam grav tCair msEspand

-- =========================================================================
-- Altera a pontuacao do jogo atual
setPontuacao :: Jogo -> Int -> Jogo
setPontuacao (Jogo t h troc ger lvl li _ c estaJ tam grav tCair msEspand) pontuacaoNova = Jogo t h troc ger lvl li pontuacaoNova c estaJ tam grav tCair msEspand

-- =========================================================================
-- Altera o combo do jogo atual
setCombo :: Jogo -> Int -> Jogo
setCombo (Jogo t h troc ger lvl li po _ estaJ tam grav tCair msEspand) comboNovo = Jogo t h troc ger lvl li po comboNovo estaJ tam grav tCair msEspand

-- =========================================================================
-- Altera o nivel do jogo atual
setNivel :: Jogo -> Int -> Jogo
setNivel (Jogo t h troc ger _ li po c estaJ tam grav tCair msEspand) lvlNovo = Jogo t h troc ger lvlNovo li po c estaJ tam grav tCair msEspand

-- =========================================================================
-- Altera a gravidade do jogo atual
setGravidade :: Jogo -> Float -> Jogo
setGravidade (Jogo t h troc ger lvl li po c estaJ tam _ tCair msEspand) gravNova = Jogo t h troc ger lvl li po c estaJ tam gravNova tCair msEspand

-- =========================================================================
-- Altera a quantidade de linhas feitas do jogo atual
setLinhas :: Jogo -> Int -> Jogo
setLinhas (Jogo t h troc ger lvl _ po c estaJ tam grav tCair msEspand) novasLinhas = Jogo t h troc ger lvl novasLinhas po c estaJ tam grav tCair msEspand

-- =========================================================================
-- Altera a tempo que ficou esperando do jogo atual
setTEsperando :: Jogo -> Float -> Jogo
setTEsperando (Jogo t h troc ger lvl li po c estaJ tam grav tCair _) tEsperandNovo = Jogo t h troc ger lvl li po c estaJ tam grav tCair tEsperandNovo

-- =========================================================================
-- Altera o tabuleiro do jogo
setTabuleiro :: Jogo -> Tabuleiro -> Jogo
setTabuleiro (Jogo _ h troc ger lvl li pont c estaJ tam grav tCair msEspand) novoTab = Jogo novoTab h troc ger lvl li pont c estaJ tam grav tCair msEspand

-- =========================================================================
-- Altera a peca que esta em hold do jogo
setHold :: Jogo -> Tetromino -> Jogo
setHold (Jogo tab _ troc ger lvl li pont c estaJ tam grav tCair msEspand) novoHold = Jogo tab (Just novoHold) True ger lvl li pont c estaJ tam grav tCair msEspand

-- =========================================================================
-- altera a variavel trocou
setTrocou :: Jogo -> Bool -> Jogo
setTrocou (Jogo tab h _ ger lvl li pont c estaJ tam grav tCair msEspand) boolean = Jogo tab h boolean ger lvl li pont c estaJ tam grav tCair msEspand

-- =========================================================================
-- Funcao chamada quando o jogador faz uma ou mais linhas. Aumenta a pontuacao
-- atual. Nessa versao a pontuacao sera aplicada depois do incremento no level
aplicaPontuacao :: Int -> Jogo -> Jogo
aplicaPontuacao qtdLinhas joguito = do
    let lvl = _level joguito
    let combo = _combo joguito
    let pontuacaoAtual = _pontuacao joguito

    let pontuacaoLinhas = avaliaPontuacao qtdLinhas lvl -- Pontuacao obtida pelas linhas limpadas
    let pontuacaoCombo  = 50*combo*lvl                  -- Pontuacao obtida pelo combo

    let pontuacaoNova = pontuacaoAtual + pontuacaoLinhas + pontuacaoCombo

    setPontuacao joguito pontuacaoNova

avaliaPontuacao :: Int -> Int -> Int
avaliaPontuacao lvl qtd
    | qtd == 1 = 100*lvl
    | qtd == 2 = 300*lvl
    | qtd == 3 = 500*lvl
    | qtd == 4 = 800*lvl
    | otherwise = 0

-- =========================================================================
-- Inicializa o jogo
inicializaJogo :: Tabuleiro -> GeradorTetros -> Float -> Jogo
inicializaJogo tabuleiro gerador tamanhoBlocos = do
    let lvl    = 1
    let linhas = 0
    let pontos = 0
    let combo  = 0
    let jogando = False
    let gravidade = gravidadeNoNivel lvl
    let tParaCair = 1/(gravidade*60)
    let tEsperando = 0

    let (gerador', tetro) = consomeTetromino gerador
    let tabuleiro' = setTetromino tabuleiro tetro

    Jogo tabuleiro' Nothing False gerador' lvl linhas pontos combo jogando tamanhoBlocos gravidade tParaCair tEsperando

-- =========================================================================
-- Muda o estado do jogo com base no tempo passado
proxEstado :: Float -> Jogo -> Jogo
proxEstado _ j@(Jogo _ _ _ _ _ _ _ _ False _ _ _ _) = j
proxEstado tempo joguito = do
    let tPCair   = _tParaCair joguito            -- Tempo que demora para o tetromino cair uma celula
    let tPassado = (_tEsperando joguito) + tempo -- Tempo passado desde a ultima vez que o tetromino caiu
    let vaiCair  = tPassado >= tPCair            -- Indica se o tetromino vai ou nao cair

    let joguitoTNovo = setTEsperando joguito tPassado -- Novo jogo com o tempo esperado corrigido

    if   vaiCair
    then derrubaTetromino joguitoTNovo
    else joguitoTNovo

-- =========================================================================
-- Move o tetromino para baixo enquanto tEsperando >= tParaCair
derrubaTetromino :: Jogo -> Jogo
derrubaTetromino joguito = do
    let tPassado = _tEsperando joguito -- Tempo passado desde a ultima vez que o tetromino caiu
    let tPCair   = _tParaCair joguito  -- Tempo que demora para o tetromino cair uma celula
    
    let (tab', caiu)   = moveTetroBaixo $ _tabuleiro joguito       -- Tabuleiro novo e indicacao se o tetromino caiu uma celula ou nao
    let joguitoTNovo   = setTEsperando joguito (tPassado - tPCair) -- Jogo com o tempo atualizado
    let joguitoTabNovo = setTabuleiro joguitoTNovo tab'            -- Jogo com o tabuleiro atualizado
    
    if   not caiu
    then mandaTetrominoNovo joguitoTNovo           -- Se nao caiu ==> chegou no final, portanto deve integrar o tetro no tabuleiro e pegar um novo
    else if (_tEsperando joguitoTabNovo) >= tPCair -- Caso contrario, avalia se dve cair ainda mais
    then derrubaTetromino joguitoTabNovo 
    else joguitoTabNovo                            -- Caso contrario, develve o jogo com o tabuleiro/tempo atualizados

-- =========================================================================
-- Funcao chamada quando um tetromino nao pode mais cair, ele eh integrado
-- no tabuleiro e um novo tetromino eh gerado
mandaTetrominoNovo :: Jogo -> Jogo
mandaTetrominoNovo joguito = do
    let gerador = _geradorTets joguito -- gerador de tetrominoes do jogo

    let (novoGerador, novoTetro)       = consomeTetromino gerador              -- gerador apos gerar o tetromino novo
    let tabuleiroTetroIntegrado        = integraTetromino (_tabuleiro joguito) -- integrando o tetromino atual
    let (tabuleiroLimpo, linhasLimpas) = limpaLinhas tabuleiroTetroIntegrado   -- limpando as linhas

    let tabuleiroFinal = setTetromino tabuleiroLimpo novoTetro -- Tabuleiro com o novo tetromino
    let contJogo       = tabuleiroFinal /= tabuleiroLimpo      -- Caso o tabuleiro final seja diferente do tabuleiro limpo, o novo tetromino nao pode ser
                                                               -- inserido nele, isso so acontece quando o meio do tabuleiro esta cheio, ou seja, o jogador
                                                               -- perdeu o jogo

    let joguitoGerNovo  = setGerador joguito novoGerador          -- Atualiza o gerador do jogo novo
    let joguitoLvlNovo  = atualizaLvl joguitoGerNovo linhasLimpas -- Atualiza o nivel do jogo novo
    let joguitoBoolNovo = setEstaJogando joguitoLvlNovo contJogo  -- Atualiza o boolean que diz se o jogo deve ou nao continuar

    let joguitoFinal   = setTabuleiro joguitoBoolNovo tabuleiroFinal               -- Atualiza o tabuleiro do jogo
    let comboNovo      = if linhasLimpas > 0 then (_combo joguitoFinal) + 1 else 0 -- Aumenta o combo, caso alguma linha tenha sido limpa

    validaJogo $ setTrocou (aplicaPontuacao linhasLimpas joguitoFinal) False -- Zera o bool que diz se o jogador pode usar a funcao de hold, aplica a pontuacao
                                                                             -- e verifica se as posicoes do spawn do novo tetromino estao livres

-- =========================================================================
-- Verifica se o jogo pode ou nao continuar com base no tabuleiro atual
validaJogo :: Jogo -> Jogo
validaJogo j = do
    let matrizTabuleiro = _grid $ _tabuleiro j              -- Grid do tabuleiro atual
    let (_, proxTetro)  = consomeTetromino $ _geradorTets j -- Novo tetromino que vai aparecer
    let larguraTetro    = larguraMatriz $ _layout proxTetro -- Largura do tetromino novo

    let xSpawn        = _x $ posInicial                 -- x da posicao inicial do spawn do tetromino
    let posicoesSpawn = [xSpawn..(xSpawn+larguraTetro)] -- Aodos os x do spawn do tetromino

    setEstaJogando j (all (avaliaGrid matrizTabuleiro) posicoesSpawn) -- Atualiza se o jogo deve continuar ou nao
        where
            avaliaGrid m x = valorEm m 0 x == Nothing

-- =========================================================================
-- Funcao chamada quando um tetromino nao pode mais cair, ele eh integrado
-- no tabuleiro e um novo tetromino eh gerado
atualizaLvl :: Jogo -> Int -> Jogo
atualizaLvl joguito novasLinhas = do
    let lvl    = _level joguito        -- Nivel atual
    let linhas = _linhasFeitas joguito -- Linhas feitas desde a ultima vez que o jogador fez 10 linhas
    let gravidade = _gravidade joguito -- gravidade atual

    let lvlNovo     = if linhas + novasLinhas >= 10 then lvl+1 else lvl                                -- Novo nivel
    let linhasNovas = if linhas + novasLinhas >= 10 then linhas+novasLinhas-10 else linhas+novasLinhas -- Quantidade de linhas limpas somando as atuais

    let gravidadeNova = if lvlNovo /= lvl then gravidadeNoNivel lvlNovo else gravidade -- Atualizando a gravidade caso o nivel tenha mudado

    let joguitoGravidadeNova = setGravidade joguito gravidadeNova    -- Jogo com a gravidade atualizada
    let joguitoLvlNovo       = setNivel joguitoGravidadeNova lvlNovo -- Jogo com o nivel atualizado
    
    setLinhas joguitoLvlNovo linhasNovas -- Jogo com as linhas feitas atualizadas

-- =========================================================================
-- Funcao que recebe e trata os inputs do jogador
eventHandler :: Event -> Jogo -> Jogo
eventHandler (EventMotion _) j = j
eventHandler (EventResize _) j = j
eventHandler (EventKey _ Up _ _) j = j
eventHandler (EventKey key _ _ _) j = trataKeyPress key j

-- Faz o tratamento correto para cada comando do usuario
trataKeyPress :: Key -> Jogo -> Jogo
trataKeyPress (MouseButton _) j = j
trataKeyPress (SpecialKey key) j
    | key == KeyLeft  = aplicaComando "leftMov" j
    | key == KeyDown  = aplicaComando "downMov" j
    | key == KeyRight = aplicaComando "rightMov" j
    | key == KeySpace = aplicaComando "hardDrop" j
    | otherwise       = j
trataKeyPress (Char key) j
    | key == 'w' = aplicaComando "hold" j
    | key == 'q' = aplicaComando "leftRot" j
    | key == 'e' = aplicaComando "rightRot" j
    | otherwise  = j

-- Transforma o input tratado em um comando para o jogo
aplicaComando :: String -> Jogo -> Jogo
-- Caso a barra de espaco seja apertada enquanto o jogo ainda nao comecou, comece ele.
-- Ingnore todos os comandos fora esse enquanto o jogo nao comecou
aplicaComando "hardDrop" j@(Jogo _ _ _ _ _ _ _ _ False _ _ _ _) = validaJogo j 
aplicaComando _ j@(Jogo _ _ _ _ _ _ _ _ False _ _ _ _)          = j
-- Os tres proximos comandos sao especiais no sentido que so fazem uma acao.
aplicaComando "hold" j     = trocaHold j
aplicaComando "downMov" j  = setTabuleiro j ((\(t, _) -> t) $ moveTetroBaixo $ _tabuleiro j)
aplicaComando "hardDrop" j = do
    let tabNovo = hardDrop $ _tabuleiro j
    let jogoNovo = setTabuleiro j tabNovo

    mandaTetrominoNovo jogoNovo
-- Os comandos que ainda nao foram tratados tem um comportamento diferente, eles zeram o tEsperando
-- caso a peca esteja no ponto mais baixo que ela pode estar. Com isso o jogador consegue transladar
-- ou rotacionar uma peca o quanto ele queira caso ela ja esteja no fundo do tabuleiro.
aplicaComando cmd j = do
    let funcaoMov = cmd2funcao cmd

    let tabuleiroAtual = _tabuleiro j
    let (_, podeCair)      = moveTetroBaixo $ tabuleiroAtual

    let (novoTabuleiro, _) = funcaoMov tabuleiroAtual
    let novoJogo           = setTabuleiro j novoTabuleiro

    if   podeCair
    then novoJogo
    else setTEsperando novoJogo 0

-- Retorna a funcao associada a string do comando
cmd2funcao :: String -> (Tabuleiro -> (Tabuleiro, Bool))
cmd2funcao "leftMov"  = moveTetroEsq
cmd2funcao "rightMov" = moveTetroDir
cmd2funcao "leftRot"  = aplicaRotacaoEsq
cmd2funcao "rightRot" = aplicaRotacaoDir

-- Troca o hold com a peca atual, o jogador ainda nao tenha feito isso com a peca mais recente
trocaHold :: Jogo -> Jogo
trocaHold j@(Jogo _ _ True _ _ _ _ _ _ _ _ _ _ ) = j
trocaHold j = do
    let holdAtual  = _hold j                   -- Tetro que esta atualmente no hold
    let tetroAtual = _tetromino $ _tabuleiro j -- Tetro que esta no tabuleiro
    let tabAtual   = _tabuleiro j              -- Tabuleiro atual

    let (ger', tet) = consomeTetromino $ _geradorTets j -- Novo gerador e novo tetromino

    let holdNovo      = if tetroAtual == Nothing then holdAtual else montaHold tetroAtual -- So muda o hold caso exista algum tetromino
    let tetroNovo     = if holdAtual == Nothing then tet else fromJust holdAtual          -- Muda o tetro atual de acordo com o conteudo do hold
    let tabuleiroNovo = setTetromino tabAtual tetroNovo                                   -- Tabuleiro com o tetromino atualizado

    let jogoTabuleiroNovo = setTabuleiro j tabuleiroNovo                  -- Jogo com o tabuleiro atualizado
    let jogoHoldNovo      = setHold jogoTabuleiroNovo (fromJust holdNovo) -- Jogo com o hold atualizado

    if   holdAtual == Nothing 
    then setGerador jogoHoldNovo ger' -- Muda o gerador caso tenha pego um novo tetromino dele
    else jogoHoldNovo

    where
        fromJust (Just x)    = x
        montaHold (Just tet) = Just $ Tetromino (layoutZero tet) (posInicial) (OZero)

-- Sempre que um tetromino vai pra zona de hold, ele reseta sua posicao e sua rotacao
layoutZero :: Tetromino -> Matriz (Maybe Bloco)
layoutZero tet 
    | tet == tetroI = Matriz layoutI
    | tet == tetroJ = Matriz layoutJ
    | tet == tetroL = Matriz layoutL
    | tet == tetroO = Matriz layoutO
    | tet == tetroS = Matriz layoutS
    | tet == tetroT = Matriz layoutT
    | tet == tetroZ = Matriz layoutZ

-- =========================================================================
-- Desenha a pontuacao atual do jogador
desenhaPontuacao :: Jogo -> Picture
desenhaPontuacao joguito = do
    let tamBloco = (_tamanhoBloco joguito)/2
    let pontos   = _pontuacao joguito
    let lvl      = _level joguito

    -- Desenhando o texto e seu background
    let background = montaRetangulo 0 (-4) (32*tamBloco) (-2*tamBloco-4) 4 white white
    let texto      = Translate 0 3 $ Scale (tamBloco/100) (tamBloco/100) $ text ("Nivel: " ++ (show lvl) ++ " | Pontuacao: " ++ (show pontos))

    Pictures [background, texto]

-- =========================================================================
-- Desenha todos os elementos do jogo
desenhaJogo :: Jogo -> (Jogo, Picture)
desenhaJogo joguito = do
    -- Tamanho de cada bloco
    let tamBloco   = _tamanhoBloco joguito

    -- Bloco do hold e seu layout
    let hold       = _hold joguito
    let holdLayout = if hold /= Nothing then _layout $ fromJust hold else Matriz layoutEmpty

    -- Desenho do Tabuleiro
    let graficoTabuleiro = desenhaTabuleiro tamBloco (_tabuleiro joguito) 

    -- Desenho do Hold. O translate coloca o desenho ao lado do tabuleiro
    -- Aqui eh feito um -2 para compensar a borda dos desenhos
    let graficoHold = Translate (-3*tamBloco - 2) (-(1.5*tamBloco/2)+2) $ desenhaHold (holdLayout) (tamBloco/2)

    -- Desenho da lista de proximos tetrominoes.
    let (gerador', graficoProximos) = desenhaProxTetros (_geradorTets joguito) (tamBloco/2)
    --  O translate coloca o desenho ao lado do tabuleiro. Aqui eh feito um +2 para compensar a borda dos desenhos
    let graficoProximosFinal        = Translate (10*tamBloco + 2) (-(1.5*tamBloco/2)+2) $ graficoProximos

    -- Jogo com o gerador atualizado
    let novoJogo = setGerador joguito gerador'

    -- Pontuacao do jogo
    let graficoPontuacao = Translate (-3*tamBloco) (0) $ desenhaPontuacao novoJogo

    -- Grafico final do jogo. O translate eh feito pois os graficos sao todos desenhados comecando do 0 0
    let graficoFinal = Translate (-5*tamBloco) (10*tamBloco) $ Pictures [graficoTabuleiro, graficoHold, graficoProximosFinal, graficoPontuacao]

    (novoJogo, graficoFinal)
        where
            fromJust (Just x) = x

-- =========================================================================
-- "Tabela" de como encontrar a gravidade com base no nivel atual
gravidadeNoNivel :: Int -> Float
gravidadeNoNivel lvl 
    | lvl == 1  = 0.01667
    | lvl == 2  = 0.021017
    | lvl == 3  = 0.026977
    | lvl == 4  = 0.035256
    | lvl == 5  = 0.04693
    | lvl == 6  = 0.06361
    | lvl == 7  = 0.0879
    | lvl == 8  = 0.1236
    | lvl == 9  = 0.1775
    | lvl == 10 = 0.2598
    | lvl == 11 = 0.388
    | lvl == 12 = 0.59
    | lvl == 13 = 0.92
    | lvl == 14 = 1.46
    | otherwise = 2.36